from __future__ import annotations

import json
import logging
import os
import pathlib
import pwd
import re
import shutil
import signal
import socket
import subprocess
import sys
import typing as t
import uuid

import daemon
import daemon.pidfile
import psutil
import setproctitle
import texttable
import yaml
from globus_compute_endpoint import __version__
from globus_compute_endpoint.endpoint.config import Config
from globus_compute_endpoint.endpoint.config.utils import serialize_config
from globus_compute_endpoint.endpoint.interchange import EndpointInterchange
from globus_compute_endpoint.endpoint.result_store import ResultStore
from globus_compute_endpoint.endpoint.utils import _redact_url_creds
from globus_compute_endpoint.logging_config import setup_logging
from globus_compute_sdk.sdk.client import Client
from globus_sdk import AuthAPIError, GlobusAPIError, NetworkError

log = logging.getLogger(__name__)


class Endpoint:
    """
    Endpoint is primarily responsible for configuring, launching and stopping
    the Endpoint.
    """

    def __init__(self, debug=False):
        """Initialize the Endpoint

        Parameters
        ----------
        debug: Bool
            Enable debug logging. Default: False
        """
        self.debug = debug

    @staticmethod
    def _config_file_path(endpoint_dir: pathlib.Path) -> pathlib.Path:
        return endpoint_dir / "config.yaml"

    @staticmethod
    def user_config_template_path(endpoint_dir: pathlib.Path) -> pathlib.Path:
        return endpoint_dir / "user_config_template.yaml"

    @staticmethod
    def user_config_schema_path(endpoint_dir: pathlib.Path) -> pathlib.Path:
        return endpoint_dir / "user_config_schema.json"

    @staticmethod
    def _user_environment_path(endpoint_dir: pathlib.Path) -> pathlib.Path:
        return endpoint_dir / "user_environment.yaml"

    @staticmethod
    def update_config_file(
        original_path: pathlib.Path,
        target_path: pathlib.Path,
        multi_tenant: bool,
        display_name: str | None,
    ):
        config_text = original_path.read_text()
        config_dict = yaml.safe_load(config_text)

        if display_name:
            config_dict["display_name"] = display_name

        if multi_tenant:
            config_dict["multi_tenant"] = multi_tenant

        config_text = yaml.safe_dump(config_dict)
        target_path.write_text(config_text)

    @staticmethod
    def init_endpoint_dir(
        endpoint_dir: pathlib.Path,
        endpoint_config: pathlib.Path | None = None,
        multi_tenant=False,
        display_name: str | None = None,
    ):
        """Initialize a clean endpoint dir

        :param endpoint_dir: Path to the endpoint configuration dir
        :param endpoint_config: Path to a config file to be used instead of
            the Globus Compute default config file
        :param multi_tenant: Whether the endpoint is a multi-user endpoint
        :param display_name: A display name to use, if desired
        """
        log.debug(f"Creating endpoint dir {endpoint_dir}")
        user_umask = os.umask(0o0077)
        os.umask(0o0077 | (user_umask & 0o0400))  # honor only the UR bit for dirs
        try:
            # pathlib.Path does not handle unusual umasks (e.g., 0o0111) so well
            # in the parents=True case, so temporarily change it.  This is nominally
            # only an issue for totally new users (no .globus_compute/!), but that is
            # also precisely the interaction -- the first one -- that should go smoothly
            endpoint_dir.mkdir(parents=True, exist_ok=True)

            config_target_path = Endpoint._config_file_path(endpoint_dir)
            package_dir = pathlib.Path(__file__).resolve().parent

            if endpoint_config is None:
                endpoint_config = package_dir / "config/default_config.yaml"

            Endpoint.update_config_file(
                endpoint_config,
                config_target_path,
                multi_tenant,
                display_name,
            )

            if multi_tenant:
                # template must be readable by user-endpoint processes (see
                # endpoint_manager.py)
                world_readable = 0o0644 & ((0o0777 - user_umask) | 0o0444)
                world_executable = 0o0711 & ((0o0777 - user_umask) | 0o0111)
                endpoint_dir.chmod(world_executable)

                src_user_tmpl_path = package_dir / "config/user_config_template.yaml"
                src_user_schem_path = package_dir / "config/user_config_schema.json"
                src_user_env_path = package_dir / "config/user_environment.yaml"
                dst_user_tmpl_path = Endpoint.user_config_template_path(endpoint_dir)
                dst_user_schem_path = Endpoint.user_config_schema_path(endpoint_dir)
                dst_user_env_path = Endpoint._user_environment_path(endpoint_dir)

                shutil.copy(src_user_tmpl_path, dst_user_tmpl_path)
                shutil.copy(src_user_schem_path, dst_user_schem_path)
                shutil.copy(src_user_env_path, dst_user_env_path)

                dst_user_tmpl_path.chmod(world_readable)
                dst_user_schem_path.chmod(world_readable)
                dst_user_env_path.chmod(world_readable)

        finally:
            os.umask(user_umask)

    @staticmethod
    def configure_endpoint(
        conf_dir: pathlib.Path,
        endpoint_config: str | None,
        multi_tenant: bool = False,
        display_name: str | None = None,
    ):
        ep_name = conf_dir.name
        if conf_dir.exists():
            print(conf_dir)
            print(f"config dir <{ep_name}> already exists")
            raise Exception("ConfigExists")

        templ_conf_path = pathlib.Path(endpoint_config) if endpoint_config else None
        Endpoint.init_endpoint_dir(
            conf_dir, templ_conf_path, multi_tenant, display_name
        )
        config_path = Endpoint._config_file_path(conf_dir)
        if multi_tenant:
            user_conf_tmpl_path = Endpoint.user_config_template_path(conf_dir)
            user_env_path = Endpoint._user_environment_path(conf_dir)

            print(f"Created multi-tenant profile for endpoint named <{ep_name}>")
            print(
                f"\n\tConfiguration file: {config_path}\n"
                f"\n\tUser endpoint configuration template: {user_conf_tmpl_path}"
                f"\n\tUser endpoint environment variables: {user_env_path}"
                "\n\nUse the `start` subcommand to run it:\n"
                f"\n\t$ globus-compute-endpoint start {ep_name}"
            )

        else:
            print(f"Created profile for endpoint named <{ep_name}>")
            print(
                f"\n\tConfiguration file: {config_path}\n"
                "\nUse the `start` subcommand to run it:\n"
                f"\n\t$ globus-compute-endpoint start {ep_name}"
            )

    @staticmethod
    def validate_endpoint_name(path_name: str) -> None:
        # Validate the path by removing all relative path modifiers (i.e., ../),
        # taking the basename, ensuring it's not Unix hidden (initial dot [.]), and
        # has no whitespace.  The path name is used deep within Parsl, which uses
        # raw shell to perform some actions ... and is not robust about it at all.
        # The path of least resistence, then, is to restrict this EP name upfront.
        # Examples:
        #
        # Good: "nice_normal_name"
        # Good: "91239anothervalidname"
        # Bad: " has_a_space"
        # Bad: "has_a_space "
        # Bad: "has a_space"
        # Bad: "has/../relative_path_syntax"
        # Bad: ".has_an_initial_dot(.);_'hidden'_directories_disallowed"
        # Bad: "a" * 129  # too long
        # Bad: ""  # come on now
        # Bad: "/absolute_path"
        # Bad: "contains...\r\v\t\n...other_whitespace"
        # Bad: "we_\\_do_not_accept_escapes"
        # Bad: "we_don't_accept_single_qoutes"
        # Bad: 'we_do_not_accept_"double_qoutes"'
        err_msg = "Invalid endpoint name: "
        pathname_re = re.compile(r"[\\/\s'\"]")
        if not path_name:
            err_msg += "Received no endpoint name"
            raise ValueError(err_msg)

        if len(path_name) > 128:
            err_msg += f"must be less than 129 characters (length: {len(path_name)})"
            raise ValueError(err_msg)

        # let pathlib resolve to an absolute (no ../), then take basename
        validated = pathlib.Path(path_name).resolve().name
        if validated != path_name:
            err_msg += (
                "no '..' or '/' characters"
                f"\n  Requested:  {path_name}"
                f"\n  Reduced to: {validated} (*potentially* valid)"
            )
            raise ValueError(err_msg)

        validated = pathname_re.sub("", validated).lstrip(".")
        if validated != path_name:
            err_msg += (
                "no whitespace (spaces, newlines, tabs), slashes, or prefixed '.'"
                f"\n  Requested:  {path_name}"
                f"\n  Reduced to: {validated} (*potentially* valid)"
            )
            raise ValueError(err_msg)

        # verify it's still a valid pathname *after* our ministrations
        validated = pathlib.Path(validated).name

        # only valid if completely unchanged by pathlib and basename shenanigans
        if path_name != validated:
            err_msg += (
                "unknown reason"
                f"\n  Requested:  {path_name}"
                f"\n  Reduced to: {validated} (*potentially* valid)"
            )
            raise ValueError(err_msg)

    @staticmethod
    def get_endpoint_id(endpoint_dir: pathlib.Path) -> str | None:
        info_file_path = endpoint_dir / "endpoint.json"
        try:
            return json.loads(info_file_path.read_bytes())["endpoint_id"]
        except FileNotFoundError:
            pass
        return None

    @staticmethod
    def get_endpoint_dir_by_uuid(
        gc_conf_dir: pathlib.Path, uuid: str
    ) -> pathlib.Path | None:
        for ep_path in Endpoint._get_ep_dirs(gc_conf_dir):
            if uuid == Endpoint.get_endpoint_id(ep_path):
                return ep_path
        return None

    @staticmethod
    def get_or_create_endpoint_uuid(
        endpoint_dir: pathlib.Path, endpoint_uuid: str | None
    ) -> str:
        ep_id = Endpoint.get_endpoint_id(endpoint_dir)
        if not ep_id:
            ep_id = endpoint_uuid or str(uuid.uuid4())
        return ep_id

    @staticmethod
    def get_funcx_client(config: Config | None) -> Client:
        if config:
            funcx_client_options = {
                "funcx_service_address": config.funcx_service_address,
                "environment": config.environment,
            }

            return Client(**funcx_client_options)
        else:
            return Client()

    def start_endpoint(
        self,
        endpoint_dir,
        endpoint_uuid,
        endpoint_config: Config,
        log_to_console: bool,
        no_color: bool,
        reg_info: dict,
        die_with_parent: bool = False,
    ):
        # This is to ensure that at least 1 executor is defined
        if not endpoint_config.executors:
            raise Exception(
                f"Endpoint config file at {endpoint_dir} is missing "
                "executor definitions"
            )

        pid_check = Endpoint.check_pidfile(endpoint_dir)
        # if the pidfile exists, we should return early because we don't
        # want to attempt to create a new daemon when one is already
        # potentially running with the existing pidfile
        if pid_check["exists"]:
            if pid_check["active"]:
                endpoint_name = endpoint_dir.name
                if endpoint_config.display_name:
                    endpoint_name = endpoint_config.display_name
                active_msg = f"Endpoint '{endpoint_name}' is already active"
                print(active_msg)
                log.info(active_msg)
                sys.exit(-1)
            else:
                log.info(
                    "A prior Endpoint instance appears to have been terminated without "
                    "proper cleanup. Cleaning up now."
                )
                Endpoint.pidfile_cleanup(endpoint_dir)

        if endpoint_config.endpoint_setup:
            Endpoint._run_command("endpoint_setup", endpoint_config.endpoint_setup)

        result_store = ResultStore(endpoint_dir=endpoint_dir)

        # Create a daemon context
        # If we are running a full detached daemon then we will send the output to
        # log files, otherwise we can piggy back on our stdout
        if endpoint_config.detach_endpoint:
            stdout: t.TextIO = open(
                os.path.join(endpoint_dir, endpoint_config.stdout), "a+"
            )
            stderr: t.TextIO = open(
                os.path.join(endpoint_dir, endpoint_config.stderr), "a+"
            )
        else:
            stdout = sys.stdout
            stderr = sys.stderr

        try:
            pid_file = endpoint_dir / "daemon.pid"
            context = daemon.DaemonContext(
                working_directory=endpoint_dir,
                umask=0o002,
                pidfile=daemon.pidfile.PIDLockFile(pid_file),
                stdout=stdout,
                stderr=stderr,
                detach_process=endpoint_config.detach_endpoint,
            )

        except Exception:
            log.exception(
                "Caught exception while trying to setup endpoint context dirs"
            )
            exit(-1)

        # place registration after everything else so that the endpoint will
        # only be registered if everything else has been set up successfully
        if not reg_info:
            endpoint_uuid = Endpoint.get_or_create_endpoint_uuid(
                endpoint_dir, endpoint_uuid
            )
            log.debug("Attempting registration; trying with eid: %s", endpoint_uuid)
            try:
                fx_client = Endpoint.get_funcx_client(endpoint_config)
                reg_info = fx_client.register_endpoint(
                    endpoint_dir.name,
                    endpoint_uuid,
                    metadata=Endpoint.get_metadata(endpoint_config),
                    multi_tenant=False,
                    display_name=endpoint_config.display_name,
                    allowed_functions=endpoint_config.allowed_functions,
                )

            except GlobusAPIError as e:
                if e.http_status in (409, 410, 423):
                    # CONFLICT, GONE or LOCKED
                    blocked_msg = f"Endpoint registration blocked.  [{e.text}]"
                    print(blocked_msg)
                    log.warning(blocked_msg)
                    exit(os.EX_UNAVAILABLE)
                raise

            except NetworkError as e:
                # the output of a NetworkError exception is huge and unhelpful, so
                # it seems better to just stringify it here and get a concise error
                log.exception(
                    f"Caught exception while attempting endpoint registration: {e}"
                )
                msg = (
                    "globus-compute-endpoint is unable to reach the Globus Compute "
                    "service due to a network error.\n"
                    "Please ensure the Globus Compute service address is reachable, "
                    "then attempt restarting the endpoint."
                )
                print(msg)
                log.critical(msg)
                exit(os.EX_TEMPFAIL)

            ret_ep_uuid = reg_info.get("endpoint_id")
            if ret_ep_uuid != endpoint_uuid:
                log.error(
                    "Unexpected response from server: mismatched endpoint id."
                    f"\n  Expected: {endpoint_uuid}, received: {ret_ep_uuid}"
                )
                exit(os.EX_SOFTWARE)

        try:
            endpoint_uuid = reg_info["endpoint_id"]
        except KeyError:
            log.error("Invalid credential structure")
            exit(os.EX_DATAERR)

        # sanitize passwords in logs
        log_reg_info = _redact_url_creds(repr(reg_info))
        log.debug(f"Registration information: {log_reg_info}")

        json_file = endpoint_dir / "endpoint.json"

        # `endpoint_id` key kept for backward compatibility when
        # globus-compute-endpoint list is called
        ep_info = {"endpoint_id": endpoint_uuid}
        json_file.write_text(json.dumps(ep_info))
        log.debug(f"Registration info written to {json_file}")

        ptitle = f"Globus Compute Endpoint ({endpoint_uuid}, {endpoint_dir.name})"
        if endpoint_config.environment:
            ptitle += f" - {endpoint_config.environment}"
        ptitle += f" [{setproctitle.getproctitle()}]"
        setproctitle.setproctitle(ptitle)

        parent_pid = 0
        if die_with_parent:
            parent_pid = os.getppid()

        log.debug("Launching endpoint daemon process")

        # NOTE
        # It's important that this log is emitted before we enter the daemon context
        # because daemonization closes down everything, a log message inside the
        # context won't write the currently configured loggers
        logfile = endpoint_dir / "endpoint.log"
        log.debug(
            "Reconfiguring logging for daemonization. logfile: %s , debug: %s",
            logfile,
            self.debug,
        )

        ostream = None
        if sys.stdout.isatty():
            ostream = sys.stdout
        elif sys.stderr.isatty():
            ostream = sys.stderr
        if ostream:
            msg = f"Starting endpoint; registered ID: {endpoint_uuid}"
            if log_to_console:
                # make more prominent against other drab gray text ...
                msg = f"\n    {msg}\n"
            print(msg, file=ostream)

        with context:
            # Per DaemonContext implementation, and that we _don't_ pass stdin,
            # fd 0 is already connected to devnull.  Unfortunately, there is an
            # as-yet unknown interaction on Polaris (ALCF) that needs this
            # connection setup *again*.  So, repeat what daemon context already
            # did, and dup2 stdin from devnull to ... devnull.  (!@#$%^&*)
            # On any other system, this should make no difference (same file!)
            with open(os.devnull) as nullf:
                if os.dup2(nullf.fileno(), 0) != 0:
                    raise Exception("Unable to close stdin")

            setup_logging(
                logfile=logfile,
                debug=self.debug,
                console_enabled=log_to_console,
                no_color=no_color,
            )

            Endpoint.daemon_launch(
                endpoint_uuid,
                endpoint_dir,
                endpoint_config,
                reg_info,
                result_store,
                parent_pid,
            )

    @staticmethod
    def daemon_launch(
        endpoint_uuid,
        endpoint_dir,
        endpoint_config: Config,
        reg_info,
        result_store: ResultStore,
        parent_pid: int,
    ):
        log.info(f"\n\n========== Endpoint begins: {endpoint_uuid}")

        interchange = EndpointInterchange(
            config=endpoint_config,
            reg_info=reg_info,
            endpoint_id=endpoint_uuid,
            endpoint_dir=endpoint_dir,
            result_store=result_store,
            logdir=endpoint_dir,
            parent_pid=parent_pid,
        )

        interchange.start()

        log.info(f"\n---------- Endpoint shutdown: {endpoint_uuid}\n")

    @staticmethod
    def stop_endpoint(
        endpoint_dir: pathlib.Path,
        endpoint_config: Config | None,
        remote: bool = False,
    ):
        pid_path = endpoint_dir / "daemon.pid"
        ep_name = endpoint_dir.name

        if remote is True:
            endpoint_id = Endpoint.get_endpoint_id(endpoint_dir)
            if not endpoint_id:
                raise ValueError(f"Endpoint <{ep_name}> could not be located")

            fx_client = Endpoint.get_funcx_client(endpoint_config)
            fx_client.stop_endpoint(endpoint_id)

        ep_status = Endpoint.check_pidfile(endpoint_dir)
        if ep_status["exists"] and not ep_status["active"]:
            Endpoint.pidfile_cleanup(endpoint_dir)
            return
        elif not ep_status["exists"]:
            log.info(f"Endpoint <{ep_name}> is not active.")
            return

        log.debug(f"{ep_name} has a daemon.pid file")

        if endpoint_config and endpoint_config.endpoint_teardown:
            Endpoint._run_command(
                "endpoint_teardown", endpoint_config.endpoint_teardown
            )

        pid = int(pid_path.read_text().strip())
        try:
            log.debug(f"Signaling process: {pid}")
            # For all the processes, including the deamon and its descendants,
            # send SIGTERM, wait for 10s, and then SIGKILL any still alive.
            grace_period_s = 10
            parent = psutil.Process(pid)
            processes = parent.children(recursive=True)
            processes.append(parent)
            for p in processes:
                p.send_signal(signal.SIGTERM)
            terminated, alive = psutil.wait_procs(processes, timeout=grace_period_s)
            for p in alive:
                try:
                    p.send_signal(signal.SIGKILL)
                except psutil.NoSuchProcess:
                    pass

            if pid_path.exists():
                log.warning(f"Endpoint <{ep_name}> did not gracefully shutdown")
                # Do cleanup anyway, TODO figure out where it should have done that
                pid_path.unlink(missing_ok=True)
            else:
                log.info(f"Endpoint <{ep_name}> is now stopped")
        except OSError:
            log.warning(f"Endpoint <{ep_name}> could not be terminated")
            log.warning(f"Attempting Endpoint <{ep_name}> cleanup")
            pid_path.unlink(missing_ok=True)
            exit(-1)

    @staticmethod
    def delete_endpoint(
        endpoint_dir: pathlib.Path, endpoint_config: Config | None, force: bool = False
    ):
        ep_name = endpoint_dir.name

        if not endpoint_dir.exists():
            log.warning(f"Endpoint <{ep_name}> does not exist")
            exit(-1)

        endpoint_id = Endpoint.get_endpoint_id(endpoint_dir)
        if endpoint_id is None:
            log.warning(
                f"Configuration for endpoint <{ep_name}> could not be found, it "
                "might not have been initialized locally"
            )
            if not force:
                exit(-1)

        # Delete endpoint from web service
        try:
            fx_client = Endpoint.get_funcx_client(endpoint_config)
            fx_client.delete_endpoint(endpoint_id)
            log.info(f"Endpoint <{ep_name}> has been deleted from the web service")
        except AuthAPIError:
            # Send to the exception handler
            raise
        except GlobusAPIError as e:
            log.warning(
                f"Endpoint <{ep_name}> could not be deleted from the web service"
                f"  [{e.text}]"
            )
            if not force:
                log.critical("Exiting without deleting the endpoint")
                exit(os.EX_UNAVAILABLE)
        except NetworkError as e:
            log.warning(
                f"Endpoint <{ep_name}> could not be deleted from the web service"
                f"  [{e}]"
            )
            if not force:
                msg = (
                    "globus-compute-endpoint is unable to reach the Globus Compute "
                    "service due to a network error.\n"
                    "Please ensure the Globus Compute service address is reachable, "
                    "then attempt to delete the endpoint again."
                )
                print(msg)
                log.critical(msg)
                exit(os.EX_TEMPFAIL)

        # Stop endpoint to handle process cleanup
        Endpoint.stop_endpoint(endpoint_dir, None)

        # Delete endpoint directory
        shutil.rmtree(endpoint_dir)

        log.info(f"Endpoint <{ep_name}> has been deleted.")

    @staticmethod
    def check_pidfile(endpoint_dir: pathlib.Path):
        """Helper function to identify possible dead endpoints

        Returns a record with 'exists' and 'active' fields indicating
        whether the pidfile exists, and whether the process is active if it does exist
        (The endpoint can only start correctly if there is no pidfile)

        Parameters
        ----------
        endpoint_dir : pathlib.Path
            Configuration directory of the endpoint
        """
        status = {"exists": False, "active": False}
        pid_path = endpoint_dir / "daemon.pid"
        if not pid_path.exists():
            return status

        status["exists"] = True

        try:
            pid = int(pid_path.read_text().strip())
            psutil.Process(pid)
            status["active"] = True
        except ValueError:
            # Invalid literal for int() parsing
            pass
        except psutil.NoSuchProcess:
            pass

        return status

    @staticmethod
    def pidfile_cleanup(endpoint_dir: pathlib.Path):
        (endpoint_dir / "daemon.pid").unlink(missing_ok=True)
        log.info(f"Endpoint <{endpoint_dir.name}> has been cleaned up.")

    @staticmethod
    def get_endpoints(funcx_conf_dir: pathlib.Path | str) -> dict[str, dict]:
        """
        Gets a dictionary that contains information about all locally
        known endpoints.

        "status" can be one of:
            ["Initialized", "Running", "Disconnected", "Stopped"]

        Example output:
        {
            "default": {
                 "status": "Running",
                 "id": "123abcde-a393-4456-8de5-123456789abc"
            },
            "my_test_ep": {
                 "status": "Disconnected",
                 "id": "xxxxxxxx-xxxx-1234-abcd-xxxxxxxxxxxx"
            }
        }
        """
        ep_statuses = {}

        for ep_path in Endpoint._get_ep_dirs(pathlib.Path(funcx_conf_dir)):
            ep_status = {
                "status": "Initialized",
                "id": Endpoint.get_endpoint_id(ep_path),
            }
            ep_statuses[ep_path.name] = ep_status
            if not ep_status["id"]:
                continue

            pid_check = Endpoint.check_pidfile(ep_path)
            if pid_check["active"]:
                ep_status["status"] = "Running"
            elif pid_check["exists"]:
                ep_status["status"] = "Disconnected"
            else:
                ep_status["status"] = "Stopped"

        return ep_statuses

    @staticmethod
    def get_running_endpoints(conf_dir: pathlib.Path):
        return {
            ep_name: ep_status
            for ep_name, ep_status in Endpoint.get_endpoints(conf_dir).items()
            if ep_status["status"].lower() == "running"
        }

    @staticmethod
    def print_endpoint_table(conf_dir: pathlib.Path, ofile=None):
        """
        Converts locally configured endpoint list to a text based table
        and prints the output.
          For example format, see the texttable module
        """
        if not ofile:
            ofile = sys.stdout

        endpoints = Endpoint.get_endpoints(conf_dir)
        if not endpoints:
            print(
                "No endpoints configured!\n\n "
                "(Hint: globus-compute-endpoint configure)",
                file=ofile,
            )
            return

        table = texttable.Texttable()
        headings = ["Endpoint ID", "Status", "Endpoint Name"]
        table.header(headings)

        idx_id = 0
        idx_st = 1
        idx_name = 2
        width_st = len(headings[idx_st])
        width_id = len(headings[idx_id])
        for ep_name, ep_info in endpoints.items():
            table.add_row([ep_info["id"], ep_info["status"], ep_name])
            width_id = max(width_id, len(str(ep_info["id"])))
            width_st = max(width_st, len(str(ep_info["status"])))

        tsize = shutil.get_terminal_size()
        max_width = max(68, tsize.columns)  # require at least a reasonable size ...
        table.set_max_width(max_width)  # ... but allow to expand to terminal width

        # trickery here, but don't want to subclass texttable -- a very stable
        # library -- at this time; ensure that the only "fungible" column is the
        # name.  Can redress ~when~ *if* this becomes a problem.  "Simple"
        table._compute_cols_width()
        if table._width[idx_id] < width_id:
            table._width[idx_name] -= width_id - table._width[idx_id]
            table._width[idx_id] = width_id
        if table._width[idx_st] < width_st:
            table._width[idx_name] -= width_st - table._width[idx_st]
            table._width[idx_st] = width_st

        # ensure no mistakes -- width of name (well, _any_) column can't be < 0
        table._width[idx_name] = max(10, table._width[idx_name])

        print(table.draw(), file=ofile)

    @staticmethod
    def get_metadata(config: Config) -> dict:
        metadata: dict = {
            "endpoint_version": __version__,
            "hostname": socket.getfqdn(),
            # should be more accurate than `getpass.getuser()` in non-login situations
            "local_user": pwd.getpwuid(os.getuid()).pw_name,
        }

        try:
            metadata["config"] = serialize_config(config)
        except Exception as e:
            log.warning(
                f"Error when serializing config ({type(e).__name__}). Ignoring."
            )
            log.debug("Config serialization exception details", exc_info=e)

        return metadata

    @staticmethod
    def _run_command(name: str, command: str):
        log.info(f"running {name} command: {command}")
        completed_process = subprocess.run(command, shell=True, capture_output=True)

        log.info(f"{name} stdout: {str(completed_process.stdout)}")
        log.info(f"{name} stderr: {str(completed_process.stderr)}")

        returncode = completed_process.returncode
        if returncode:
            log.error(f"{name} failed with exit code {returncode}")
            exit(os.EX_CONFIG)

    @staticmethod
    def _get_ep_dirs(gc_conf_dir: pathlib.Path) -> t.Iterable[pathlib.Path]:
        return (p.parent for p in gc_conf_dir.glob("*/config.*"))



from __future__ import annotations

import logging
import multiprocessing
import os
import platform
import queue
import random
import signal
import sys
import threading
import time
from concurrent.futures import Future

# multiprocessing.Event is a method, not a class
# to annotate, we need the "real" class
# see: https://github.com/python/typeshed/issues/4266
from multiprocessing.synchronize import Event as EventType

import globus_compute_endpoint.endpoint.config
import pika.exceptions
import setproctitle
from globus_compute_common.messagepack import InvalidMessageError, pack
from globus_compute_common.messagepack.message_types import (
    EPStatusReport,
    Result,
    ResultErrorDetails,
)
from globus_compute_endpoint import __version__ as funcx_endpoint_version
from globus_compute_endpoint.endpoint.rabbit_mq import (
    ResultPublisher,
    TaskQueueSubscriber,
)
from globus_compute_endpoint.endpoint.result_store import ResultStore
from globus_compute_endpoint.engines.base import GlobusComputeEngineBase
from globus_compute_endpoint.exception_handling import get_result_error_details
from globus_compute_sdk import __version__ as funcx_sdk_version
from parsl.version import VERSION as PARSL_VERSION

log = logging.getLogger(__name__)


class EndpointInterchange:
    """Interchange is a task orchestrator for distributed systems.

    1. Asynchronously queue large volume of tasks (>100K)
    2. Allow for workers to join and leave the union
    3. Detect workers that have failed using heartbeats
    4. Service single and batch requests from workers
    5. Be aware of requests worker resource capacity,
       eg. schedule only jobs that fit into walltime.

    """

    def __init__(
        self,
        config: globus_compute_endpoint.endpoint.config.Config,
        reg_info: dict[str, dict],
        logdir=".",
        endpoint_id=None,
        endpoint_dir=".",
        result_store: ResultStore | None = None,
        reconnect_attempt_limit: int = 5,
        parent_pid: int = 0,
    ):
        """
        Parameters
        ----------
        config : globus_compute_sdk.Config object
             Globus Compute config object describing how compute should be provisioned

        reg_info : dict[str, dict]
             Dictionary containing connection information for both the task and
             result queues.  The required data structure is returned from the
             Endpoint registration API call, encapsulated in the SDK by
             `Client.register_endpoint()`.

        logdir : str
             Parsl log directory paths. Logs and temp files go here. Default: '.'

        endpoint_id : str
             Identity string that identifies the endpoint to the broker

        endpoint_dir : pathlib.Path
             Endpoint directory path to store registration info in
        """
        self.logdir = logdir
        log.info(
            "Initializing EndpointInterchange process with Endpoint ID: {}".format(
                endpoint_id
            )
        )
        self.config = config

        self.endpoint_dir = endpoint_dir

        self.task_q_info = reg_info["task_queue_info"]
        self.result_q_info = reg_info["result_queue_info"]

        self.time_to_quit = False
        self.heartbeat_period = self.config.heartbeat_period

        self.pending_task_queue: multiprocessing.Queue = multiprocessing.Queue()

        self._reconnect_fail_counter = 0
        self.reconnect_attempt_limit = max(1, reconnect_attempt_limit)
        self._quiesce_event = multiprocessing.Event()
        self._kill_event = multiprocessing.Event()
        self._parent_pid = parent_pid

        if result_store is None:
            result_store = ResultStore(endpoint_dir=endpoint_dir)
        self.result_store = result_store

        self.endpoint_id = endpoint_id

        self.current_platform = {
            "parsl_v": PARSL_VERSION,
            "python_v": "{}.{}.{}".format(
                sys.version_info.major, sys.version_info.minor, sys.version_info.micro
            ),
            "os": platform.system(),
            "hname": platform.node(),
            "funcx_sdk_version": funcx_sdk_version,
            "funcx_endpoint_version": funcx_endpoint_version,
            "registration": self.endpoint_id,
            "dir": os.getcwd(),
        }
        log.info(f"Platform info: {self.current_platform}")

        self.results_passthrough: queue.Queue[
            dict[str, bytes | str | None]
        ] = queue.Queue()
        # Rename self.executor -> self.engine in second round
        self.executor: GlobusComputeEngineBase = self.config.executors[0]
        self._test_start = False

    def start_engine(self):
        log.info("Starting Engine")
        self.executor.start(
            results_passthrough=self.results_passthrough,
            endpoint_id=self.endpoint_id,
            run_dir=self.logdir,
        )

    def migrate_tasks_to_internal(
        self,
        connection_params: dict,
        endpoint_uuid: str,
        pending_task_queue: multiprocessing.Queue,
        quiesce_event: EventType,
    ) -> multiprocessing.Process:
        """Pull tasks from the incoming tasks 0mq pipe onto the internal
        pending task queue

        Parameters:
        -----------
        connection_params: pika.connection.Parameters
              Connection params to connect to the service side Tasks queue

        endpoint_uuid: endpoint_uuid str

        pending_task_queue: multiprocessing.Queue
              Internal queue to which tasks should be migrated

        quiesce_event : EventType
              Event to let the thread know when it is time to die.
        """
        try:
            log.info(f"Starting the TaskQueueSubscriber as {endpoint_uuid}")
            task_q_proc = TaskQueueSubscriber(
                queue_info=connection_params,
                external_queue=pending_task_queue,
                quiesce_event=quiesce_event,
                endpoint_id=endpoint_uuid,
            )
            task_q_proc.start()
        except Exception:
            log.exception("Unhandled exception in TaskQueueSubscriber")
            raise

        return task_q_proc

    def quiesce(self):
        """Temporarily stop everything on the interchange in order to reach a consistent
        state before attempting to start again. This must be called on the main thread
        """
        log.info("Interchange Quiesce in progress (stopping and joining processes)")
        self._quiesce_event.set()

        log.info("Waiting for quiesce complete")
        self._task_puller_proc.join()

        log.info("Quiesce done")

    def stop(self):
        """Prepare the interchange for shutdown"""
        log.info("Shutting down EndpointInterchange")

        # kill_event must be set before quiesce_event because we need to guarantee that
        # once the quiesce is complete, the interchange will not try to start again
        self._kill_event.set()
        self._quiesce_event.set()

    def cleanup(self):
        self.executor.shutdown(block=True)

    def handle_sigterm(self, sig_num, curr_stack_frame):
        log.warning("Received SIGTERM, setting termination flag.")
        self.time_to_quit = True

    def function_allowed(self, function_id: str):
        if self.config.allowed_functions is None:
            # Not a restricted endpoint
            return True

        return function_id in self.config.allowed_functions

    def start(self):
        """Start the Interchange"""
        signal.signal(signal.SIGHUP, self.handle_sigterm)
        signal.signal(signal.SIGQUIT, self.handle_sigterm)  # hint: Ctrl+\
        signal.signal(signal.SIGTERM, self.handle_sigterm)
        # Intentionally ignoring SIGINT for now, as we're unstable enough to
        # warrant Python's default developer-friendly Ctrl+C handling

        self._quiesce_event.clear()
        self._kill_event.clear()

        if self._parent_pid:
            if self._parent_pid != os.getppid():
                # initial check to save extra noise
                log.warning(f"Pid {self._parent_pid} not parent; refusing to start")
                self.stop()
                return

            def _parent_watcher(ppid: int):
                while ppid == os.getppid():
                    if not self._quiesce_event.wait(timeout=1):
                        return
                log.warning(f"Parent ({ppid}) has gone away; initiating shut down")
                self.stop()

            threading.Thread(
                target=_parent_watcher,
                args=(self._parent_pid,),  # copy; to discourage shenanigans
                daemon=True,
                name="ParentPidWatcher",
            ).start()

        log.info("Starting EndpointInterchange")

        # NOTE: currently we only start the executors once because
        # the current behavior is to keep them running decoupled while
        # the endpoint is waiting for reconnection
        self.start_engine()

        while not self._kill_event.is_set():
            if self._reconnect_fail_counter >= self.reconnect_attempt_limit:
                log.critical(
                    f"Failed {self._reconnect_fail_counter} consecutive times."
                    "  Shutting down."
                )
                self.stop()
                self.quiesce()
                break

            if self._quiesce_event.is_set():
                idle_for = random.uniform(2.0, 10.0)
                log.warning(f"Interchange will retry connecting in {idle_for:.2f}s")
                time.sleep(idle_for)
                self._quiesce_event.clear()
            else:
                log.debug("Starting threads and main loop")

            if self.time_to_quit:
                self.stop()
                self.quiesce()
                break

            try:
                self._start_threads_and_main()
            except pika.exceptions.ConnectionClosedByBroker as e:
                log.warning(f"AMQP service closed connection: {e}")
            except pika.exceptions.ProbableAuthenticationError as e:
                log.error(f"Unable to connect to AMQP service: {e}")
                self._kill_event.set()
            except Exception:
                log.error("Unhandled exception in main kernel.")
                log.debug("Unhandled exception in main kernel.", exc_info=True)
            finally:
                if not self._kill_event.is_set():
                    self._reconnect_fail_counter += 1
                    log.info(
                        "Reconnection count: %s (of %s)",
                        self._reconnect_fail_counter,
                        self.reconnect_attempt_limit,
                    )

            self.quiesce()

        self.cleanup()
        log.info("EndpointInterchange shutdown complete.")

    def _start_threads_and_main(self):
        self._task_puller_proc = self.migrate_tasks_to_internal(
            self.task_q_info,
            self.endpoint_id,
            self.pending_task_queue,
            self._quiesce_event,
        )

        self._main_loop()

    def _main_loop(self):
        """
        This is the "kernel" of the endpoint interchange process.  Conceptually, there
        are three actions of consequence: forward task messages to the executors,
        forward results from the executors back to the Globus Compute web services
        (RMQ), and forward any previous results that may have failed to send previously
        (e.g., if a RMQ connection was dropped).

        We accomplish this via three threads, one each for each task.

        Of special note is that this kernel does not try very hard to handle the
        non-happy path.  If an error occurs that is too "unhappy" (e.g., communication
        with RMQ fails), the catch-all solution is to "reboot" and try again.  That is,
        this method exits, communication is shutdown, started again, and we restart
        this method.
        """
        log.debug("_main_loop begins")

        results_publisher = ResultPublisher(queue_info=self.result_q_info)
        results_publisher.start()

        executor = self.executor

        num_tasks_forwarded = 0
        num_results_forwarded = 0

        def process_stored_results():
            # Handle any previously stored results, either from a previous run or
            # from a quarter of a second-ago.  Basically, check every second
            # (.wait()) if there are any items on disk to be sent.  If there are,
            # don't treat them any differently than "fresh" results: put the into
            # the same multiprocessing queue as results incoming directly from
            # the executors.  The normal processing by `process_pending_results()`
            # will take over from there.
            while not self._quiesce_event.wait(timeout=1):
                for task_id, packed_result in self.result_store:
                    log.debug("Retrieved stored result (%s)", task_id)
                    msg = {"task_id": task_id, "message": packed_result}
                    self.result_store.discard(task_id)
                    self.results_passthrough.put(msg)

                    if self._quiesce_event.is_set():
                        # important to check every iteration as well, so as not to
                        # potentially hang up the shutdown procedure
                        return
            log.debug("Exit process-stored-results thread.")

        def process_pending_tasks():
            # Pull tasks from upstream (RMQ) and send them down the ZMQ pipe to the
            # globus-compute-manager.  In terms of shutting down (or "rebooting")
            # gracefully, iterate once a second whether a task has arrived.
            nonlocal num_tasks_forwarded
            while not self._quiesce_event.is_set():
                if self.time_to_quit:
                    self.stop()
                    continue  # nominally == break; but let event do it

                try:
                    prop_headers, body = self.pending_task_queue.get(timeout=1)

                    fid: str = prop_headers.get("function_uuid")
                    tid: str = prop_headers.get("task_uuid")
                    if not fid or not tid:
                        raise InvalidMessageError("Task message missing headers")

                    if fid and not self.function_allowed(fid):
                        # Same as web-service message but packed in a
                        # result error
                        reject_msg = (
                            f"Function {fid} not permitted on "
                            f"endpoint {self.endpoint_id}"
                        )
                        log.warning(reject_msg)

                        failed_result = Result(
                            task_id=tid,
                            data=reject_msg,
                            error_details=ResultErrorDetails(
                                code="FUNCTION_NOT_ALLOWED", user_message=reject_msg
                            ),
                        )
                        results_publisher.publish(pack(failed_result))
                        continue

                except queue.Empty:
                    continue

                except Exception:
                    log.exception("Unhandled error processing incoming task")
                    continue

                try:
                    executor.submit(task_id=tid, packed_task=body)
                    num_tasks_forwarded += 1  # Safe given GIL

                except Exception as exc:
                    log.exception(f"Failed to process task {tid}")
                    code, msg = get_result_error_details()
                    failed_result = Result(
                        task_id=tid,
                        data=f"Failed to start task: {exc}",
                        error_details=ResultErrorDetails(code=code, user_message=msg),
                        task_statuses=[],
                    )
                    res = {
                        "task_id": tid,
                        "message": pack(failed_result),
                    }
                    self.results_passthrough.put(res)

            log.debug("Exit process-pending-tasks thread.")

        def process_pending_results():
            # Forward incoming results from the globus-compute-manager to the
            # Globus Compute services. For graceful handling of shutdown
            # (or "reboot"), wait up to a second or incoming results before
            # iterating the loop regardless.
            nonlocal num_results_forwarded

            def _create_done_cb(mq_msg: bytes, tid: str | None):
                def _done_cb(pub_fut: Future):
                    _exc = pub_fut.exception()
                    if _exc:
                        # Publishing didn't work -- quiesce and see if a simple
                        # restart fixes the issue.
                        if tid:
                            log.info(f"Storing result for later: {tid}")
                            self.result_store[tid] = mq_msg

                        self._quiesce_event.set()
                        log.error("Failed to publish results", exc_info=_exc)

                return _done_cb

            while not self._quiesce_event.is_set():
                try:
                    msg = self.results_passthrough.get(timeout=1)
                    packed_message: bytes = msg["message"]
                    task_id: str | None = msg.get("task_id")

                except queue.Empty:
                    continue

                except Exception as exc:
                    log.warning(
                        "Invalid message received.  Ignoring."
                        f"  ([{type(exc).__name__}] {exc})"
                    )
                    continue

                if task_id:
                    num_results_forwarded += 1
                    log.debug("Forwarding result for task: %s", task_id)

                try:
                    f = results_publisher.publish(packed_message)
                    f.add_done_callback(_create_done_cb(packed_message, task_id))

                except Exception:
                    # Publishing didn't work -- quiesce and see if a simple restart
                    # fixes the issue.
                    self._quiesce_event.set()

                    log.exception("Something broke while forwarding results")
                    if task_id:
                        log.info("Storing result for later: %s", task_id)
                        self.result_store[task_id] = packed_message
                    continue  # just be explicit

            log.debug("Exit process-pending-results thread.")

        stored_processor_thread = threading.Thread(
            target=process_stored_results, name="Stored Result Handler"
        )
        task_processor_thread = threading.Thread(
            target=process_pending_tasks, name="Pending Task Handler"
        )
        result_processor_thread = threading.Thread(
            target=process_pending_results, name="Pending Result Handler"
        )
        stored_processor_thread.start()
        task_processor_thread.start()
        result_processor_thread.start()

        last_t, last_r = 0, 0

        soft_idle_limit = max(0, self.config.idle_heartbeats_soft)
        hard_idle_limit = max(soft_idle_limit + 1, self.config.idle_heartbeats_hard)
        soft_idle_heartbeats = 0  # "happy path" idle timeout
        hard_idle_heartbeats = 0  # catch-all idle timeout

        live_proc_title = setproctitle.getproctitle()
        log.debug("_main_loop entered running state")
        while not self._quiesce_event.wait(self.heartbeat_period):
            # Possibly TOCTOU here, but we don't need to be super precise.  The
            # point here is to mention "still alive" and that we're still working
            num_t, num_r = num_tasks_forwarded, num_results_forwarded
            diff_t, diff_r = num_t - last_t, num_r - last_r
            log.debug(
                "Heartbeat.  Approximate Tasks and Results forwarded since last "
                "heartbeat: %s (T), %s (R)",
                diff_t,
                diff_r,
            )
            last_t, last_r = num_t, num_r

            # only reset come heartbeat and still alive
            self._reconnect_fail_counter = 0

            if not soft_idle_limit:
                # idle timeout not enabled; "always on"
                continue

            if diff_t or diff_r:
                # a task moved; reset idle heartbeat counter
                if soft_idle_heartbeats or hard_idle_heartbeats:
                    log.info(
                        "Moved to active state (due to tasks processed since"
                        " last heartbeat)."
                    )
                setproctitle.setproctitle(live_proc_title)
                soft_idle_heartbeats = 0
                hard_idle_heartbeats = 0
                continue

            # only start "timer" if we've at least done *some* work
            hard_idle_heartbeats += 1
            if (num_t or num_r) and num_r >= num_t:
                # similar to above, only start "timer" if *idle* ... but
                # note that given self.result_store, it's possible to
                # have forwarded more results than tasks received.
                soft_idle_heartbeats += 1
                shutdown_s = soft_idle_limit - soft_idle_heartbeats
                shutdown_s *= self.heartbeat_period

                if soft_idle_heartbeats == 1:
                    log.info(
                        "In idle state (due to no task or result movement);"
                        f" shut down in {shutdown_s:,}s.  (idle_heartbeats_soft)"
                    )
                idle_proc_title = "[idle; shut down in {:,}s] {}"
                setproctitle.setproctitle(
                    idle_proc_title.format(shutdown_s, live_proc_title)
                )

                if soft_idle_heartbeats >= soft_idle_limit:
                    log.info("Idle heartbeats reached.  Shutting down.")
                    self.time_to_quit = True
                    self.stop()

            elif hard_idle_heartbeats > hard_idle_limit:
                log.warning("Shutting down due to idle heartbeats HARD limit.")
                self.time_to_quit = True
                self.stop()

            elif hard_idle_heartbeats > soft_idle_limit:
                # Reminder: this branch only hit if EP started and no tasks
                # or results have moved.  If *any* movement occurs, this branch
                # won't get executed.
                shutdown_s = hard_idle_limit - hard_idle_heartbeats
                shutdown_s *= self.heartbeat_period
                if hard_idle_heartbeats == soft_idle_limit + 1:
                    # only log the first time; no sense in filling logs
                    log.info(
                        "Possibly idle -- no task or result movement.  Will"
                        f" shut down in {shutdown_s:,}s.  (idle_heartbeats_hard)"
                    )
                idle_proc_title = "[possibly idle; shut down in {:,}s] {}"
                setproctitle.setproctitle(
                    idle_proc_title.format(shutdown_s, live_proc_title)
                )

        # The timeouts aren't nominally necessary because if the above loop has
        # quit, then the _quiesce_event is set, and both threads check that event
        # every internal iteration.  But "for kicks."
        stored_processor_thread.join(timeout=5)
        task_processor_thread.join(timeout=5)
        result_processor_thread.join(timeout=5)

        # let higher-level error handling take over if the following excepts
        message = EPStatusReport(
            endpoint_id=self.endpoint_id,
            global_state={
                "managers": 0,
                "total_workers": 0,
                "idle_workers": 0,
                "pending_tasks": 0,
                "outstanding_tasks": {},
                "heartbeat_period": 0,  # 0 == "shutting down now"
            },
            task_statuses={},
        )
        f = results_publisher.publish(pack(message))
        f.result(timeout=5)
        results_publisher.stop()

        log.debug("_main_loop exits")



from __future__ import annotations

import logging
import multiprocessing
import queue
import typing as t
import uuid
from concurrent.futures import Future
from concurrent.futures import ProcessPoolExecutor as NativeExecutor

import psutil
from globus_compute_common.messagepack.message_types import (
    EPStatusReport,
    TaskTransition,
)
from globus_compute_endpoint.engines.base import (
    GlobusComputeEngineBase,
    ReportingThread,
)

logger = logging.getLogger(__name__)


class ProcessPoolEngine(GlobusComputeEngineBase):
    def __init__(
        self,
        *args,
        label: str = "ProcessPoolEngine",
        heartbeat_period_s: float = 30.0,
        **kwargs,
    ):
        self.label = label
        self.executor: t.Optional[NativeExecutor] = None
        self._executor_args = args
        self._executor_kwargs = kwargs
        self._status_report_thread = ReportingThread(
            target=self.report_status, args=[], reporting_period=heartbeat_period_s
        )
        super().__init__(*args, heartbeat_period_s=heartbeat_period_s, **kwargs)

    def start(
        self,
        *args,
        endpoint_id: t.Optional[uuid.UUID] = None,
        results_passthrough: t.Optional[queue.Queue] = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        endpoint_id: Endpoint UUID
        results_passthrough: Queue to which packed results will be posted
        Returns
        -------
        """
        if self.executor is None:
            # We are instantiating the executor here, rather than in the constructor,
            # to ensure the executor starts within the daemon context. Doing so avoids
            # having the daemon close existing pipe file descriptors required for
            # inter-process communication.
            self.executor = NativeExecutor(
                *self._executor_args, **self._executor_kwargs
            )

        assert endpoint_id, "ProcessPoolExecutor requires kwarg:endpoint_id at start"
        self.endpoint_id = endpoint_id
        if results_passthrough:
            self.results_passthrough = results_passthrough
        assert self.results_passthrough

        self._status_report_thread.start()

    def get_status_report(self) -> EPStatusReport:
        assert self.executor, "The engine has not been started"
        executor_status: t.Dict[str, t.Any] = {
            "task_id": -2,
            "info": {
                "total_cores": multiprocessing.cpu_count(),
                "total_mem": round(psutil.virtual_memory().available / (2**30), 1),
                "total_core_hrs": 0,
                "total_workers": self.executor._max_workers,  # type: ignore
                "pending_tasks": 0,
                "outstanding_tasks": 0,
                "scaling_enabled": False,
                "max_blocks": 1,
                "min_blocks": 1,
                "max_workers_per_node": self.executor._max_workers,  # type: ignore
                "nodes_per_block": 1,
                "heartbeat_period": self._heartbeat_period_s,
            },
        }
        task_status_deltas: t.Dict[str, t.List[TaskTransition]] = {}

        return EPStatusReport(
            endpoint_id=self.endpoint_id,
            global_state=executor_status,
            task_statuses=task_status_deltas,
        )

    def _submit(
        self,
        func: t.Callable,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> Future:
        """We pass all params except the function to executor.submit()"""
        assert self.executor, "The engine has not been started"
        logger.warning("Got task")
        return self.executor.submit(func, *args, **kwargs)

    def status_polling_interval(self) -> int:
        return 30

    def shutdown(self, /, block=False, **kwargs) -> None:
        self._status_report_thread.stop()
        if self.executor:
            self.executor.shutdown(wait=block)



from __future__ import annotations

import typing as t

import pytest
import responses
from pika.exchange_type import ExchangeType


@pytest.fixture(autouse=True)
def _autouse_responses():
    responses.start()

    yield

    responses.stop()
    responses.reset()


# mock logging config to do nothing
@pytest.fixture(autouse=True)
def _mock_logging_config(monkeypatch):
    def fake_setup_logging(*args, **kwargs):
        pass

    monkeypatch.setattr(
        "globus_compute_endpoint.logging_config.setup_logging", fake_setup_logging
    )


@pytest.fixture
def setup_register_endpoint_response(
    create_result_queue_info,
    task_queue_info,
    ensure_task_queue,
    ensure_result_queue,
) -> t.Callable[[str], None]:
    ensure_task_queue(queue_opts={"queue": task_queue_info["queue"]})

    def resp(endpoint_uuid: str):
        rq_info = create_result_queue_info(queue_id=endpoint_uuid)
        exchange_name = rq_info["exchange"]
        queue_name = rq_info["test_routing_key"]
        exchange_opts = {
            "exchange": exchange_name,
            "exchange_type": ExchangeType.topic.value,
            "durable": True,
        }
        queue_opts = {"queue": queue_name, "durable": True}
        ensure_result_queue(exchange_opts=exchange_opts, queue_opts=queue_opts)

        responses.add(
            method=responses.POST,
            url="https://compute.api.globus.org/v2/endpoints",
            headers={"Content-Type": "application/json"},
            json={
                "endpoint_id": endpoint_uuid,
                "task_queue_info": task_queue_info,
                "result_queue_info": rq_info,
            },
        )

    yield resp
